from os import makedirs, listdir, remove
from os.path import basename, join, exists
from random import random, randrange, choice, randint
from glob import glob
from array import array
from sys import argv

from pygame import PixelArray, Surface, Color, joystick, version, mixer, draw
from pygame.event import clear, Event
from pygame.image import load, save
from pygame.transform import flip, scale, smoothscale
from pygame.mixer import Sound, set_num_channels, get_num_channels, pause, unpause, stop
from pygame.key import get_mods
from pygame.locals import *

from lib.pgfw.pgfw.Game import Game
from lib.pgfw.pgfw.GameChild import GameChild
from lib.pgfw.pgfw.Animation import Animation
from lib.pgfw.pgfw.Sprite import Sprite, RainbowSprite
from lib.pgfw.pgfw.Vector import Vector
from lib.pgfw.pgfw.extension import (get_color_swapped_surface, get_delta,
                                     get_busy_channel_count)

class PictureProcessing(Game):

    SCALE = 3
    CHANNEL_COUNT = 16
    KEYBOARD, GAMEPAD = range(2)

    def __init__(self):
        Game.__init__(self)
        mixer.init(44100, -16, self.CHANNEL_COUNT, 4096)
        self.get_configuration().type_declarations.add("bool", "mouse", "resets")
        self.get_configuration().type_declarations.add("bool", "editor", "enabled")
        self.get_configuration().type_declarations.add("bool", "display", "slides-enabled")
        self.get_configuration().type_declarations.add("bool", "display", "use-joypad-text")
        self.sound_effects = SoundEffects(self)
        self.glyphs = Glyphs(self)
        self.editor = Editor(self)
        self.levels = Levels(self)
        self.interface = Interface(self)
        self.title = Title(self)
        self.introduction = Introduction(self)
        self.ending = Ending(self)
        self.slides = Slides(self)
        self.subscribe(self.respond)
        self.subscribe(self.respond, MOUSEBUTTONDOWN)
        self.reset()
        clear()
        if self.check_command_line("-wallpaper"):
            self.write_wallpaper()

    def are_slides_enabled(self):
        return self.check_command_line("-slides") or self.get_configuration("display", "slides-enabled")

    def write_wallpaper(self):
        index = argv.index("--wallpaper")
        # width, height = map(int, argv[index + 1:index + 3])
        width, height = [int(size) for size in argv[index + 1:index + 3]]
        wallpaper = ScrollingBackground(self, size=(width, height))
        surface = wallpaper.get_current_frame()
        if argv[index + 3] == "-":
            from tempfile import mkstemp
            from sys import stdout
            from os import unlink
            path = mkstemp(".png")[1]
            save(surface, path)
            # print path
        else:
            save(surface, argv[index + 3])
        self.delegate.post("quit")

    def respond(self, event):
        if self.delegate.compare(event, "reset-game") or \
           (self.get_configuration("mouse", "resets") and event.type == MOUSEBUTTONDOWN and
            (1 <= event.button <= 3)):
            self.reset()
        elif self.are_slides_enabled() and self.delegate.compare(event, "slides"):
            self.slides.toggle()
            if self.slides.active:
                if get_mods() & KMOD_LSHIFT:
                    self.slides.increment_index(-1)
                elif get_mods() & KMOD_RSHIFT:
                    self.slides.increment_index(1)

    def reset(self):
        self.title.reset()
        self.levels.reset()
        self.interface.reset()
        self.introduction.reset()
        self.ending.reset()
        self.editor.reset()
        self.slides.reset()

    def get_progress(self):
        return [int(field) for field in open(self.get_resource("progress")).read().split()]
        # return map(int, open(self.get_resource("progress")).read().split())

    def write_progress(self, level_index, swap_count, update=False):
        if update:
            swap_count += self.get_progress()[1]
        open(self.get_resource("progress"), "w").write("%s %s\n" % (str(level_index),
                                                                    str(swap_count)))

    def select_text(self, texts):
        return texts[self.get_configuration("display", "use-joypad-text")]

    def update(self):
        if not self.slides.active:
            self.editor.update()
            self.introduction.update()
            self.title.update()
            self.levels.update()
            self.interface.update()
            self.ending.update()
        else:
            self.slides.update()


class Slides(GameChild):

    ROOT = "local/egw-slides"
    VIDEOBALL_INDEX = 4

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        self.loaded = False
        self.subscribe(self.respond)
        self.index = 0

    def activate(self):
        self.active = True
        if not self.loaded:
            self.load()
        pause()
        if self.index == self.VIDEOBALL_INDEX:
            self.play_videoball()

    def deactivate(self):
        self.active = False
        self.videoball_audio.stop()
        unpause()

    def toggle(self):
        if self.active:
            self.deactivate()
        else:
            self.activate()

    def reset(self):
        self.active = False
        if self.loaded:
            self.videoball_audio.stop()

    def load(self):
        tile = self.get_game().levels.levels[1].tiles[3]
        self.background = ScrollingBackground(self, [tile], fill=(255, 0, 0))
        self.slides = []
        for ii, path in enumerate(sorted(glob(join(self.ROOT, "*")))):
            dsr = self.get_display_surface().get_rect()
            if ii == self.VIDEOBALL_INDEX:
                slide = Sprite(self)
                for frame_path in sorted(glob("local/videoball-frames/*.png")):
                    slide.add_frame(load(frame_path).convert())
                slide.location.center = dsr.center
                self.slides.append(slide)
            slide = Sprite(self)
            original_image = load(path).convert()
            original_width, original_height = original_image.get_size()
            if original_width > original_height:
                size = dsr.w, int(original_height * (float(dsr.w) / original_width))
            else:
                size = int(original_width * (float(dsr.h) / original_height)), dsr.h
            scaled_image = smoothscale(original_image, size)
            slide.add_frame(scaled_image)
            slide.location.center = dsr.center
            self.slides.append(slide)
        self.videoball_audio = Sound("local/videoball-frames/audio.wav")
        self.loaded = True

    def respond(self, event):
        if self.active:
            delegate = self.get_game().delegate
            if delegate.compare(event, "right"):
                self.increment_index(1)
            elif delegate.compare(event, "left"):
                self.increment_index(-1)

    def increment_index(self, increment):
        self.index = (self.index + increment) % len(self.slides)
        if self.index == self.VIDEOBALL_INDEX:
            self.play_videoball()
        else:
            self.videoball_audio.stop()

    def play_videoball(self):
        self.slides[self.index].get_current_frameset().reset()
        self.videoball_audio.stop()
        self.videoball_audio.play()

    def update(self):
        if self.active:
            self.background.update()
            slide = self.slides[self.index]
            if self.index == self.VIDEOBALL_INDEX:
                frameset = slide.get_current_frameset()
                if frameset.current_index == len(frameset.order) - 1:
                    self.play_videoball()
            slide.update()


class Editor(GameChild):

    TILE_SIZE = 8
    PIXEL_ZOOM = 32
    CANVAS_SCALE = 40
    LABEL_EDITOR_HEADING = "EDITOR"
    LABEL_EXIT = "ESC: EXIT", "B: EXIT"
    LABEL_EDIT = "ENTER: EDIT", "A: EDIT"
    LABEL_TILES = "\\24 EDIT TILES"
    LABEL_PLAY = "TEST \\26"
    LABEL_PAINT_ARROW = "\\25 PLACE TILES"
    LABEL_MUSIC = "\\27 MUSIC"
    LABEL_SAVE_EXIT = "ESC: SAVE", "B: SAVE"
    LABEL_HIDE_MENU = "ENTER: HIDE", "A: HIDE"
    LABEL_BACK = "ESC: BACK", "B: BACK"
    LABEL_PAINT = "Z: PAINT", "A: PAINT"
    LABEL_NEW = "ENTER: NEW", "A: NEW"
    LABEL_SELECT = "ENTER: SELECT", "A: SELECT"
    LABEL_DRAG = "HOLD Z: BOX", "HOLD A: BOX"
    LABEL_SWAP = "ENTER: SWAP", "A: SWAP"
    LABEL_SOLVED = "SOLVED"
    LABEL_LEFT = "\\27"
    LABEL_RIGHT = "\\26"
    LABEL_TILE_LOAD = "LOAD"
    LABEL_TILE_NEW = "NEW"
    LABEL_TILE_DELETE = "DELETE"
    LABEL_TILE_EDIT = "EDIT"
    LABEL_MIX_COLOR = "ENTER: MIX", "A: MIX"
    LABEL_ADD_COLOR = "ENTER: ADD", "A: ADD"
    LABEL_RED = "\\27 RED \\26"
    LABEL_GREEN = "\\27 GREEN \\26"
    LABEL_BLUE = "\\27 BLUE \\26"
    LABEL_SWITCH_TILE = "LEFT SHIFT: SWITCH TILE", "SELECT: SWITCH TILE"
    HEADING_MARGIN = 60
    BUTTONS_MARGIN = 60
    VIEW_MARGIN = 120
    DIRECTIONS_MARGIN = 0
    ALIGN_ONE_THIRD_BOTTOM, ALIGN_TWO_THIRDS_BOTTOM = range(2)
    TILE_BAR_PADDING = 10
    TILE_BAR_BACKGROUND = 0, 0, 0
    TILE_BAR_SCALE = 3
    PALETTE_WIDTH = 100
    ZFILL_FILE = 5
    TILE_MENU_CURSOR = "\\16"
    TILE_MENU_SPACING = 16
    ZFILL_TILE_PATH = 2
    TILE_LOAD_PREVIEW_SCALE = 16
    TILE_LOAD_MENU_SHRINK = -100, -300
    TILE_LOAD_MENU_PADDING = 8
    TILE_LOAD_MENU_MARGIN = 2
    BOX_PREVIEW_ALPHA = 200
    MIXER_STEP = 10

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        self.brush_position = Vector()
        self.subscribe(self.respond)
        self.levels = []
        self.painting = []
        for _ in range(self.TILE_SIZE):
            self.painting.append([0] * self.TILE_SIZE)
        self.loaded = False
        self.mixer_color = MixableColor()
        self.added_colors = []

    def reset(self):
        self.at_level_select = True
        self.at_view = False
        self.at_tile_bar = False
        self.at_tile_edit = False
        self.at_palette = False
        self.at_box = False
        self.at_test = False
        self.at_tile_menu = False
        self.at_tile_load = False
        self.at_mixer = False
        self.level_index = 0
        self.tile_index = 0
        self.palette_index = 0
        self.tile_load_index = 0
        self.mixer_index = 0
        self.tile_menu_index = 0
        self.menu_hidden = False
        self.paint_pressed = False
        self.box = FlashingCursor(self, *([self.get_level_tile_size()] * 2))
        self.deactivate()

    def activate(self):
        self.active = True
        if not self.loaded:
            self.load()
        self.paint_pressed = False
        self.set_level_title()

    def load(self):
        self.read_levels()
        glyphs = self.get_game().glyphs
        dsr = self.get_display_surface().get_rect()
        self.editor_heading = self.get_label(self.LABEL_EDITOR_HEADING, True)
        self.editor_heading.location.midtop = dsr.centerx, self.HEADING_MARGIN
        self.arrows = Sprite(self), Sprite(self)
        self.background = ScrollingBackground(self, self.levels[-1].tiles)
        self.level_select_exit = self.get_label(
            self.LABEL_EXIT, align=self.ALIGN_ONE_THIRD_BOTTOM,
            margin=self.BUTTONS_MARGIN)
        self.label_edit_level = self.get_label(
            self.LABEL_EDIT, align=self.ALIGN_TWO_THIRDS_BOTTOM,
            margin=self.BUTTONS_MARGIN)
        self.view_exit = self.get_label(
            self.LABEL_SAVE_EXIT, align=self.ALIGN_ONE_THIRD_BOTTOM,
            margin=self.VIEW_MARGIN)
        self.view_hide = self.get_label(
            self.LABEL_HIDE_MENU, align=self.ALIGN_TWO_THIRDS_BOTTOM,
            margin=self.VIEW_MARGIN)
        self.label_back = self.get_label(
            self.LABEL_BACK, align=self.ALIGN_ONE_THIRD_BOTTOM,
            margin=self.VIEW_MARGIN)
        self.label_back_box = self.get_label(
            self.LABEL_BACK, align=self.ALIGN_ONE_THIRD_BOTTOM,
            margin=self.VIEW_MARGIN)
        self.label_paint = self.get_label(
            self.LABEL_PAINT, align=self.ALIGN_TWO_THIRDS_BOTTOM,
            margin=self.VIEW_MARGIN)
        self.label_edit_view = self.get_label(
            self.LABEL_EDIT, align=self.ALIGN_TWO_THIRDS_BOTTOM,
            margin=self.VIEW_MARGIN)
        self.label_new = self.get_label(
            self.LABEL_NEW, align=self.ALIGN_TWO_THIRDS_BOTTOM,
            margin=self.VIEW_MARGIN)
        self.label_select = self.get_label(
            self.LABEL_SELECT, align=self.ALIGN_TWO_THIRDS_BOTTOM,
            margin=self.VIEW_MARGIN)
        self.label_drag = self.get_label(
            self.LABEL_DRAG, align=self.ALIGN_TWO_THIRDS_BOTTOM,
            margin=self.VIEW_MARGIN)
        self.label_left = self.get_label(self.LABEL_LEFT, True, red=200)
        self.label_left.location.midleft = dsr.left + self.BUTTONS_MARGIN, dsr.centery
        self.label_right = self.get_label(self.LABEL_RIGHT, True, red=200)
        self.label_right.location.midright = dsr.right - self.BUTTONS_MARGIN, dsr.centery
        self.label_swap = self.get_label(
            self.LABEL_SWAP, align=self.ALIGN_TWO_THIRDS_BOTTOM,
            margin=self.VIEW_MARGIN)
        self.label_mix_color = self.get_label(
            self.LABEL_MIX_COLOR, align=self.ALIGN_TWO_THIRDS_BOTTOM,
            margin=self.VIEW_MARGIN)
        self.label_add_color = self.get_label(
            self.LABEL_ADD_COLOR, align=self.ALIGN_TWO_THIRDS_BOTTOM,
            margin=self.VIEW_MARGIN)
        self.label_red = self.get_label(self.LABEL_RED)
        self.label_red.location.centerx = dsr.centerx
        self.label_green = self.get_label(self.LABEL_GREEN)
        self.label_green.location.centerx = dsr.centerx
        self.label_blue = self.get_label(self.LABEL_BLUE)
        self.label_blue.location.centerx = dsr.centerx
        self.label_solved = self.get_label(self.LABEL_SOLVED)
        self.label_solved.location.bottomright = dsr.bottomright
        self.label_tile_load = self.get_label(self.LABEL_TILE_LOAD)
        self.label_tile_edit = self.get_label(self.LABEL_TILE_EDIT)
        self.label_tile_new = self.get_label(self.LABEL_TILE_NEW)
        self.label_tile_delete = self.get_label(self.LABEL_TILE_DELETE)
        self.label_switch_tile = self.get_label(self.LABEL_SWITCH_TILE)
        self.label_switch_tile.location.centerx = dsr.centerx
        self.arrow_tiles = self.get_label(self.LABEL_TILES)
        self.arrow_tiles.location.midtop = dsr.midtop
        self.arrow_play = self.get_label(self.LABEL_PLAY)
        self.arrow_play.location.midright = dsr.midright
        self.arrow_paint = self.get_label(self.LABEL_PAINT_ARROW)
        self.arrow_paint.location.midbottom = dsr.midbottom
        self.arrow_music = self.get_label(self.LABEL_MUSIC)
        self.arrow_music.location.midleft = dsr.midleft
        self.tile_bar = Sprite(self)
        bar_tile_height = self.get_current_level().original_tiles[0].get_width() * \
            self.TILE_BAR_SCALE
        surface = Surface((dsr.w, bar_tile_height + self.TILE_BAR_PADDING * 2))
        surface.fill(self.TILE_BAR_BACKGROUND)
        self.tile_bar.add_frame(surface)
        self.tile_menu_background = Sprite(self)
        surface = Surface((dsr.w, self.label_tile_new.location.h))
        surface.fill((0, 0, 0))
        self.tile_menu_background.add_frame(surface)
        self.tile_menu_background.location.top = self.tile_bar.location.bottom
        self.label_tile_load.location.top = self.tile_menu_background.location.top
        self.label_tile_edit.location.top = self.tile_menu_background.location.top
        self.label_tile_new.location.top = self.tile_menu_background.location.top
        self.label_tile_delete.location.top = self.tile_menu_background.location.top
        self.level_select_cursor = self.get_cursor(
            *self.get_current_level().preview.get_size())
        self.level_select_cursor.location.center = dsr.center
        self.tile_cursor = self.get_cursor(
            *([self.TILE_SIZE * self.TILE_BAR_SCALE] * 2))
        self.paint_cursor = self.get_cursor(*([self.CANVAS_SCALE] * 2))
        self.palette_cursor = self.get_cursor()
        self.mixer_cursor = self.get_cursor(300, 100)
        self.mixer_cursor.location.center = dsr.center
        self.tile_menu_cursor = self.get_label(self.TILE_MENU_CURSOR)
        self.tile_menu_cursor.location.top = self.tile_menu_background.location.top
        new_tile = self.new_tile = Surface([self.TILE_SIZE] * 2)
        new_tile.fill((255, 255, 255))
        for x in range(new_tile.get_width()):
            for y in range(new_tile.get_height()):
                if 3 <= x <= 4 and y != 0 and y != 7:
                    new_tile.set_at((x, y), (0, 0, 0))
                elif 3 <= y <= 4 and x != 0 and x != 7:
                    new_tile.set_at((x, y), (0, 0, 0))
        self.build_default_palette()
        self.tile_load_background = Sprite(self)
        surface = Surface(dsr.inflate(self.TILE_LOAD_MENU_SHRINK).size)
        surface.fill((0, 0, 0))
        self.tile_load_background.add_frame(surface)
        self.tile_load_background.location.center = dsr.center
        self.tile_load_cursor = self.get_cursor(*([self.get_level_tile_size()] * 2))
        # self.tile_load_preview_background = Sprite(self)
        # rect = Rect((0, 0), [self.TILE_SIZE * self.TILE_LOAD_PREVIEW_SCALE] * 2)
        # rect.inflate_ip(16, 16)
        # surface = Surface(rect.size)
        # surface.fill((0, 0, 0))
        # self.tile_load_preview_background.add_frame(surface)
        # self.tile_load_preview_background.location.topright = self.tile_load_background.location.topright
        self.loaded = True
        self.current_level_is_default = False
        self.level_index = len(self.levels) - 1
        # for level in self.levels:
            # print "path:", level.path

    def draw_tile_database(self):
        padding = self.TILE_LOAD_MENU_PADDING
        x, y = self.get_tile_load_topleft()
        ds = self.get_display_surface()
        self.tile_database = []
        for level in self.levels + self.get_game().levels.levels:
            self.tile_database.extend(level.original_tiles)
        self.tile_database.extend(self.get_game().glyphs.original_tiles)
        count = 0
        for tile in self.tile_database:
            scaled = scale(tile, [self.get_level_tile_size()] * 2)
            ds.blit(scaled, (x, y))
            x += scaled.get_width() + self.TILE_LOAD_MENU_MARGIN
            # if y < self.tile_load_preview_background.location.bottom:
            #     limit = self.tile_load_preview_background.location.left
            # else:
            # limit = self.tile_load_background.location.right - padding
            # if x + tile.get_width() > limit:
            count += 1
            if count == self.get_tile_load_row_count():
                count = 0
                x = self.tile_load_background.location.left + padding
                y += scaled.get_height() + self.TILE_LOAD_MENU_MARGIN
        # preview = Sprite(self)
        # surface = scale(self.tile_database[self.tile_load_index],
        #                 [self.TILE_LOAD_PREVIEW_SCALE * self.TILE_SIZE] * 2)
        # preview.add_frame(surface)
        # preview.location.center = self.tile_load_preview_background.location.center
        # preview.update()

    def get_tile_load_row_count(self):
        return int((self.tile_load_background.location.w - self.TILE_LOAD_MENU_PADDING * 2) / \
            (self.get_level_tile_size() + self.TILE_LOAD_MENU_MARGIN))

    def get_tile_load_cursor_step(self):
        return self.get_level_tile_size() + self.TILE_LOAD_MENU_MARGIN

    def get_tile_load_row_width(self):
        return self.get_tile_load_row_count() * self.get_tile_load_cursor_step()

    def get_tile_load_topleft(self):
        x = self.tile_load_background.location.left + self.TILE_LOAD_MENU_PADDING
        y = self.tile_load_background.location.top + self.TILE_LOAD_MENU_PADDING
        return x, y

    def get_level_tile_size(self):
        return self.TILE_SIZE * PictureProcessing.SCALE

    def build_default_palette(self):
        self.default_palette = [Color(255, 255, 255), Color(0, 0, 0)]
        for hue in range(0, 360 - 45, 45):
            for lightness in range(25, 100, 25):
                color = Color(0, 0, 0)
                color.hsla = hue, 100, lightness, 100
                self.default_palette.append(color)

    def get_label(self, text, rainbow=False, align=None, margin=0, red=80):
        if isinstance(text, tuple):
            text = self.get_game().select_text(text)
        surface = self.get_game().glyphs.get_surface_from_text(
            text, background=([0, red][rainbow], 0, 0))
        if rainbow:
            label = RainbowSprite(self, surface)
        else:
            label = Sprite(self)
            label.add_frame(surface)
        if align is not None:
            dsr = self.get_display_surface().get_rect()
            if align == self.ALIGN_ONE_THIRD_BOTTOM:
                label.location.midbottom = dsr.centerx / 2, dsr.h - margin
            elif align == self.ALIGN_TWO_THIRDS_BOTTOM:
                label.location.midbottom = \
                    dsr.centerx / 2 + dsr.centerx, dsr.h - margin
        return label

    def read_levels(self):
        levels = self.levels = []
        for path in sorted(glob(join(self.get_new_levels_path(), "*"))):
            level = Level(self)
            if level.load(Level.LOAD_GRID_FILE, path):
                levels.append(level)
            # else:
            #     print "Failed to load level %s" % path
        self.add_default_level()

    def add_default_level(self):
        self.levels.append(self.get_default_level())

    def get_default_level(self):
        level = Level(self)
        level.load()
        return level

    def get_default_tile_path(self):
        return join(self.get_tiles_path(),
                    self.get_configuration("editor", "default-tile-file-name"))

    def get_tiles_path(self):
        config = self.get_configuration("editor")
        return join(config["root-directory"], config["tiles-directory"])

    def get_levels_path(self):
        config = self.get_configuration("editor")
        return join(config["root-directory"], config["levels-directory"])

    def get_new_levels_path(self):
        return join(self.get_levels_path(), self.get_configuration("editor", "new-directory"))

    def get_original_levels_path(self):
        return join(self.get_levels_path(), self.get_configuration("editor", "original-directory"))

    def deactivate(self):
        self.active = False
        self.get_game().interface.reset()

    def respond(self, event):
        if self.active and not self.get_game().slides.active:
            compare = self.get_delegate().compare
            dx, dy = 0, 0
            if compare(event, "action"):
                if self.at_level_select:
                    self.at_level_select = False
                    self.at_view = True
                    self.menu_hidden = False
                    if self.level_index == len(self.levels) - 1:
                        self.current_level_is_default = True
                elif self.at_view:
                    self.menu_hidden = not self.menu_hidden
                elif self.at_tile_bar:
                    self.at_tile_menu = True
                    self.at_tile_bar = False
                    self.tile_menu_index = 0
                elif self.at_tile_menu:
                    if self.tile_menu_index == 0:
                        if self.add_tile_selected():
                            tile = Surface([self.TILE_SIZE] * 2)
                            tile.fill((0, 0, 0))
                            self.get_current_level().add_tile(tile)
                            if self.current_level_is_default:
                                self.current_level_is_default = False
                                self.add_default_level()
                        self.at_tile_edit = True
                        self.at_tile_menu = False
                    elif self.tile_menu_index == 1:
                        self.tile_load_index = 0
                        self.tile_load_cursor.location.topleft = self.get_tile_load_topleft()
                        self.at_tile_load = True
                        self.at_tile_menu = False
                    elif self.tile_menu_index == 2:
                        self.delete_tile(self.tile_index)
                        self.at_tile_menu = False
                        self.at_tile_bar = True
                elif self.at_tile_load:
                    level = self.get_current_level()
                    tile = self.tile_database[self.tile_load_index].copy()
                    tile.set_colorkey(None)
                    if self.add_tile_selected():
                        level.add_tile(tile)
                    else:
                        level.original_tiles[self.tile_index] = tile
                        level.tiles[self.tile_index] = scale(tile, [tile.get_width() * PictureProcessing.SCALE] * 2)
                    if self.current_level_is_default:
                        self.current_level_is_default = False
                        self.add_default_level()
                    self.save_level()
                    level.set_preview()
                    self.at_tile_load = False
                    self.at_tile_bar = True
                elif self.at_palette:
                    if self.at_mixer:
                        self.add_color()
                        if not self.is_mixer_selected():
                            self.palette_index = len(self.get_full_palette())
                    elif self.is_mixer_selected():
                        self.at_mixer = True
                    else:
                        self.at_palette = False
                        self.at_tile_edit = True
            if compare(event, "cancel"):
                if self.at_level_select:
                    self.deactivate()
                    self.get_game().title.activate()
                elif self.at_view:
                    self.at_view = False
                    self.at_level_select = True
                    self.set_level_title()
                    self.tile_index = 0
                    self.palette_index = 0
                elif self.at_tile_bar:
                    self.at_tile_bar = False
                    self.at_view = True
                elif self.at_tile_menu:
                    self.at_tile_menu = False
                    self.at_tile_bar = True
                elif self.at_tile_load:
                    self.at_tile_load = False
                    self.at_tile_bar = True
                elif self.at_tile_edit:
                    self.at_tile_edit = False
                    self.at_tile_bar = True
                elif self.at_palette:
                    if self.at_mixer:
                        self.at_mixer = False
                    elif not self.is_mixer_selected():
                        self.at_palette = False
                        self.at_tile_edit = True
                elif self.at_box:
                    self.at_box = False
                    self.at_view = True
                elif self.at_test and self.get_game().interface.closed:
                    self.at_test = False
                    self.at_view = True
                    self.get_current_level().set_swap_status()
                    self.get_game().interface.deactivate()
            if compare(event, "right"):
                if self.at_level_select:
                    self.level_index += 1
                    if self.level_index >= len(self.levels):
                        self.level_index = 0
                    self.set_level_title()
                if self.at_tile_bar:
                    self.tile_index += 1
                    if self.tile_index == Interface.MAX_TILE_COUNT or \
                       self.tile_index > len(self.get_current_level().tiles):
                        self.tile_index = 0
                if self.at_tile_menu:
                    self.tile_menu_index += 1
                    if self.tile_menu_index > 1:
                        if len(self.get_current_level().tiles) == 1 or self.add_tile_selected():
                            self.tile_menu_index = 0
                    if self.tile_menu_index > 2:
                        self.tile_menu_index = 0
                if self.at_tile_edit or self.at_box:
                    dx = 1
                if self.at_tile_edit:
                    if self.brush_position.x == self.TILE_SIZE - 1:
                        self.at_tile_edit = False
                        self.at_palette = True
                        if self.palette_index_is_at_left():
                            self.wrap_palette_index()
                elif self.at_palette:
                    if self.at_mixer:
                        shift = [0, 0, 0]
                        shift[self.mixer_index] += self.MIXER_STEP
                        self.mixer_color.shift_rgb(*shift)
                    elif self.palette_index_is_at_left():
                        self.at_palette = False
                        self.at_tile_edit = True
                    else:
                        self.wrap_palette_index()
                if self.at_tile_load:
                    row_count = self.get_tile_load_row_count()
                    if (self.tile_load_index + 1) % row_count == 0:
                        # self.tile_load_index -= row_count - 1
                        # self.tile_load_cursor.move(dx=-self.get_tile_load_row_width())
                        # self.tile_load_cursor.move(dx=self.get_tile_load_cursor_step())
                        pass
                    elif self.tile_load_index == len(self.tile_database) - 1:
                        # self.tile_load_index = 0
                        # self.tile_load_cursor.location.topleft = self.get_tile_load_topleft()
                        pass
                    else:
                        self.tile_load_index += 1
                        self.tile_load_cursor.move(dx=self.get_tile_load_cursor_step())
                if self.at_view:
                    self.at_view = False
                    self.at_test = True
                    self.get_current_level().set_swap_status(True)
                    self.get_game().interface.setup()
            elif compare(event, "down"):
                if self.at_tile_edit or self.at_box:
                    dy = 1
                if self.at_tile_load:
                    row_count = self.get_tile_load_row_count()
                    if self.tile_load_index + row_count > len(self.tile_database) - 1:
                        pass
                    else:
                        self.tile_load_index += row_count
                        self.tile_load_cursor.move(dy=self.get_tile_load_cursor_step())
                if self.at_palette:
                    if self.at_mixer:
                        self.mixer_index = (self.mixer_index + 1) % 3
                    else:
                        self.palette_index += 1
                if self.at_view:
                    self.at_view = False
                    self.at_box = True
                    if self.add_tile_selected():
                        self.tile_index = 0
            elif compare(event, "left"):
                if self.at_level_select:
                    self.level_index -= 1
                    if self.level_index < 0:
                        self.level_index = len(self.levels) - 1
                    self.set_level_title()
                if self.at_tile_bar:
                    self.tile_index -= 1
                    if self.tile_index < 0:
                        self.tile_index = min(
                            Interface.MAX_TILE_COUNT - 1,
                            len(self.get_current_level().tiles))
                if self.at_tile_menu:
                    self.tile_menu_index -= 1
                    if self.tile_menu_index < 0:
                        if self.add_tile_selected() or len(self.get_current_level().tiles) == 1:
                            self.tile_menu_index = 1
                        else:
                            self.tile_menu_index = 2
                if self.at_tile_edit or self.at_box:
                    dx = -1
                if self.at_tile_edit:
                    if self.brush_position.x == 0:
                        self.at_tile_edit = False
                        self.at_palette = True
                        if not self.palette_index_is_at_left():
                            self.wrap_palette_index()
                elif self.at_palette:
                    if self.at_mixer:
                        shift = [0, 0, 0]
                        shift[self.mixer_index] -= self.MIXER_STEP
                        self.mixer_color.shift_rgb(*shift)
                    elif self.is_mixer_selected():
                        self.at_mixer = True
                    elif not self.palette_index_is_at_left():
                        self.at_palette = False
                        self.at_tile_edit = True
                    else:
                        self.wrap_palette_index()
                if self.at_tile_load:
                    row_count = self.get_tile_load_row_count()
                    if self.tile_load_index == 0:
                        pass
                    elif self.tile_load_index % row_count == 0:
                        pass
                    else:
                        self.tile_load_index -= 1
                        self.tile_load_cursor.move(dx=-self.get_tile_load_cursor_step())
            elif compare(event, "up"):
                if self.at_view:
                    self.at_view = False
                    self.at_tile_bar = True
                if self.at_tile_edit or self.at_box:
                    dy = -1
                if self.at_tile_load:
                    row_count = self.get_tile_load_row_count()
                    if self.tile_load_index - row_count < 0:
                        pass
                    else:
                        self.tile_load_index -= row_count
                        self.tile_load_cursor.move(dy=-self.get_tile_load_cursor_step())
                if self.at_palette:
                    if self.at_mixer:
                        self.mixer_index = (self.mixer_index - 1) % 3
                    else:
                        self.palette_index -= 1
            elif compare(event, "paint"):
                self.paint_pressed = True
                if self.at_box:
                    self.set_box_anchor()
                if self.at_tile_edit and self.palette_index < len(self.get_full_palette()):
                    self.paint_tile()
            elif compare(event, "paint", cancel=True):
                self.paint_pressed = False
                if self.at_box and not self.add_tile_selected():
                    self.stamp()
                self.box.size = Vector(*([self.get_level_tile_size()] * 2))
            elif compare(event, "select"):
                if self.at_box:
                    self.tile_index += 1
                    if self.tile_index == len(self.get_current_level().tiles):
                        self.tile_index = 0
            if dx != 0 or dy != 0:
                if self.at_box:
                    if self.paint_pressed:
                        self.grow_box(x=dx, y=dy)
                    else:
                        self.move_box(x=dx, y=dy)
                if self.at_tile_edit:
                    self.move_brush_position(dx, dy)
                    if self.paint_pressed and self.palette_index < len(self.get_full_palette()):
                        self.paint_tile()
            if self.palette_index >= self.get_palette_cell_count():
                self.palette_index = 0
            elif self.palette_index < 0:
                self.palette_index = self.get_palette_cell_count() - 1

    def add_color(self):
        palette = self.get_full_palette()
        found = False
        for color in palette:
            if color == self.mixer_color:
                found = True
                break
        if not found:
            self.added_colors.append(Color(self.mixer_color.r, self.mixer_color.g, self.mixer_color.b))

    def set_level_title(self):
        level = self.get_current_level()
        if level.path is None:
            text = "DEFAULT"
        else:
            text = basename(self.get_current_level().path)
        rect = self.get_current_level().preview.get_rect()
        rect.center = self.get_display_surface().get_rect().center
        self.label_title = self.get_label("ID: %s" % text)
        self.label_title.location.bottomright = rect.right, rect.bottom - 3

    def delete_tile(self, index):
        level = self.get_current_level()
        level.tiles.pop(index)
        level.original_tiles.pop(index)
        for ii in range(len(level.grid)):
            for jj in range(len(level.grid[ii])):
                if level.grid[ii][jj] == index:
                    level.grid[ii][jj] = 0
                elif level.grid[ii][jj] > index:
                    level.grid[ii][jj] -= 1
        if level.path is not None:
            path = join(level.path, "%s.png" % str(index).zfill(self.ZFILL_TILE_PATH))
            if exists(path):
                remove(path)
        level.set_preview()
        self.save_level()

    def stamp(self):
        size = self.get_level_tile_size()
        x = int(self.box.location.left / size)
        y = int(self.box.location.top / size)
        w = int(self.box.location.w / size)
        h = int(self.box.location.h / size)
        for xi in range(x, x + w):
            for yi in range(y, y + h):
                self.get_current_level().grid[xi][yi] = self.tile_index
        self.box.location.topleft = [size * coordinate for coordinate in self.box_corner]
        level = self.get_current_level()
        level.set_swap_status()
        level.set_preview()
        self.save_level()

    def save_level(self):
        config = self.get_configuration("editor")
        new_levels_directory = self.get_new_levels_path()
        if not exists(new_levels_directory):
            makedirs(new_levels_directory)
        grid = self.get_current_level().grid
        level = self.get_current_level()
        if level.path is None:
            directories = sorted(listdir(new_levels_directory))
            index = len(directories) - 1
            while True:
                if index < 0:
                    level.path = join(new_levels_directory, "0" * self.ZFILL_FILE)
                    break
                elif directories[index].isdigit():
                    level.path = join(
                        new_levels_directory,
                        str(int(directories[index]) + 1).zfill(self.ZFILL_FILE))
                    makedirs(level.path)
                    break
                index -= 1
        level_file = open(join(level.path, self.get_configuration("editor", "level-file-name")), "w")
        for jj in range(len(grid[0])):
            for ii in range(len(grid)):
                level_file.write(str(hex(grid[ii][jj]))[-1])
            level_file.write("\n")
        for ii, tile in enumerate(level.original_tiles):
            save(tile, join(level.path, "%s.png" % str(ii).zfill(self.ZFILL_TILE_PATH)))

    def set_box_anchor(self):
        size = self.get_level_tile_size()
        self.box_anchor = Vector(self.box.location.left / size,
                                 self.box.location.top / size)
        self.box_corner = Vector(*self.box_anchor)

    def grow_box(self, x=0, y=0):
        size = self.get_level_tile_size()
        dsr = self.get_display_surface().get_rect()
        self.box_corner.x += x
        if self.box_corner.x < 0:
            self.box_corner.x = 0
        elif self.box_corner.x >= dsr.right / size:
            self.box_corner.x = dsr.right / size - 1
        self.box_corner.y += y
        if self.box_corner.y < 0:
            self.box_corner.y = 0
        elif self.box_corner.y >= dsr.bottom / size:
            self.box_corner.y = dsr.bottom / size - 1
        self.box.size = Vector((abs(self.box_corner.x - self.box_anchor.x) + 1) * size,
                               (abs(self.box_corner.y - self.box_anchor.y) + 1) * size)
        self.box.location.topleft = \
            min(self.box_corner.x, self.box_anchor.x) * size, \
            min(self.box_corner.y, self.box_anchor.y) * size
        self.box.update()
        
    def move_box(self, x=0, y=0):
        size = self.get_level_tile_size()
        dsr = self.get_display_surface().get_rect()
        self.box.move(x * size, y * size)
        if self.box.location.right > dsr.right + self.box.thickness or \
           self.box.location.left < -self.box.thickness:
            self.box.move(dx=-x * dsr.w)
        elif self.box.location.bottom > dsr.bottom + self.box.thickness or \
             self.box.location.top < -self.box.thickness:
            self.box.move(dy=-y * dsr.h)

    def paint_tile(self):
        original = self.get_current_tile()
        scaled = self.get_current_level().tiles[self.tile_index]
        x, y = self.brush_position
        color = self.get_full_palette()[self.palette_index]
        if original.get_at((x, y)) != color:
            if self.current_level_is_default:
                self.current_level_is_default = False
                self.add_default_level()
            original.set_at((x, y), color)
            scale = PictureProcessing.SCALE
            scaled.fill(color, (x * scale, y * scale, scale, scale))
            self.get_current_level().set_preview()
            self.save_level()

    def wrap_palette_index(self):
        count = self.get_palette_cell_count()
        if count % 2 and self.palette_index == int(count / 2):
            self.palette_index = count - 1
        else:
            self.palette_index += [-1, 1][self.palette_index_is_at_left()] * \
                (int(count / 2) + count % 2)

    def get_palette_cell_count(self):
        return len(self.get_full_palette()) + 1

    def palette_index_is_at_left(self):
        length = self.get_palette_cell_count()
        return self.palette_index < int(length / 2) + length % 2

    def move_brush_position(self, dx, dy):
        self.brush_position.move(dx, dy)
        if self.brush_position.x >= self.TILE_SIZE:
            self.brush_position.x = self.TILE_SIZE - 1
        elif self.brush_position.x < 0:
            self.brush_position.x = 0
        if self.brush_position.y >= self.TILE_SIZE:
            self.brush_position.y = 0
        elif self.brush_position.y < 0:
            self.brush_position.y = self.TILE_SIZE - 1

    def get_current_level(self):
        return self.levels[self.level_index]

    def get_cursor(self, width=0, height=0, flashing=True):
        return FlashingCursor(self, width, height, flashing)

    def add_tile_selected(self):
        return self.tile_index == len(self.get_current_level().original_tiles)

    def update(self):
        if self.active:
            ds = self.get_display_surface()
            ds.fill((0, 0, 0))
            if self.at_level_select:
                self.background.update()
                self.editor_heading.update()
                self.level_select_exit.update()
                self.label_edit_level.update()
                preview = self.get_current_level().preview
                rect = preview.get_rect()
                rect.center = ds.get_rect().center
                ds.blit(preview, rect)
                self.label_title.update()
                self.level_select_cursor.update()
                if len(self.levels) > 1:
                    self.label_left.update()
                    self.label_right.update()
            else:
                self.get_current_level().update()
            if self.at_view:
                if not self.menu_hidden:
                    self.view_exit.update()
                    self.view_hide.update()
                    self.arrow_tiles.update()
                    self.arrow_play.update()
                    self.arrow_paint.update()
                    # self.arrow_music.update()
            if self.at_tile_bar or self.at_tile_edit or self.at_palette or self.at_tile_menu or self.at_tile_load:
                self.tile_bar.update()
                tiles = self.get_current_level().original_tiles
                if len(tiles) < Interface.MAX_TILE_COUNT:
                    full_tiles = tiles + [self.new_tile]
                else:
                    full_tiles = tiles
                for ii, tile in enumerate(full_tiles):
                    cell_width = self.tile_bar.location.w / Interface.MAX_TILE_COUNT
                    scaled_tile = scale(
                        tile, [tiles[0].get_width() * self.TILE_BAR_SCALE] * 2)
                    rect = scaled_tile.get_rect()
                    rect.center = cell_width * ii + cell_width / 2, \
                        self.tile_bar.location.centery
                    ds.blit(scaled_tile, rect)
                    if ii == self.tile_index:
                        self.tile_cursor.location.center = rect.center
                        if self.at_tile_bar:
                            self.tile_cursor.set_flashing(True)
                            self.tile_cursor.update()
                        else:
                            self.tile_cursor.set_flashing(False)
                self.label_back.update()
                if self.at_tile_menu:
                    self.tile_menu_background.update()
                    if not self.add_tile_selected():
                        labels = self.label_tile_edit, self.label_tile_load
                        if len(self.get_current_level().tiles) > 1:
                            labels += self.label_tile_delete,
                    else:
                        labels = self.label_tile_new, self.label_tile_load
                    width = self.tile_menu_cursor.location.w * (len(labels) - 1) + \
                        self.TILE_MENU_SPACING * (len(labels) - 1)
                    for label in labels:
                        width += label.location.w
                    x = ds.get_width() / 2 - width / 2
                    for label in labels:
                        label.location.left = x
                        x += label.location.w + self.tile_menu_cursor.location.w + self.TILE_MENU_SPACING
                        label.update()
                    self.tile_menu_cursor.location.right = labels[self.tile_menu_index].location.left
                    self.tile_menu_cursor.update()
                    self.label_select.update()
                if not self.add_tile_selected():
                    tile = tiles[self.tile_index]
                    canvas = scale(tile, [self.CANVAS_SCALE * tile.get_width()] * 2)
                    rect = canvas.get_rect()
                    rect.center = ds.get_rect().center
                    if self.is_mixer_selected():
                        surface = Surface(rect.size)
                        surface.fill(self.mixer_color)
                        ds.blit(surface, rect)
                    else:
                        ds.blit(canvas, rect)
                    palette = self.get_full_palette()
                    length = len(palette) + 1
                    color_left_height = canvas.get_height() / (length / 2 + length % 2)
                    color_right_height = canvas.get_height() / (length / 2)
                    palette_left_height = color_left_height * (length / 2 + length % 2)
                    palette_right_height = color_right_height * (length / 2)
                    palette_left_rect = Rect(0, 0, self.PALETTE_WIDTH, palette_left_height)
                    palette_left_rect.midright = rect.left - 30, rect.centery
                    palette_right_rect = Rect(0, 0, self.PALETTE_WIDTH, palette_right_height)
                    palette_right_rect.midleft = rect.right + 30, rect.centery
                    y = palette_left_rect.top
                    for ii in range(length):
                        if ii < length / 2 + length % 2:
                            palette_rect = palette_left_rect
                            h = color_left_height
                        else:
                            palette_rect = palette_right_rect
                            h = color_left_height
                        if ii < length - 1:
                            ds.fill(palette[ii], (palette_rect.left, y, palette_rect.w, h))
                        else:
                            for x in range(palette_rect.w):
                                hue = int(float(x) / palette_rect.w * 360)
                                color = Color(0, 0, 0)
                                color.hsla = hue, 100, 50, 100
                                ds.fill(color, (palette_rect.left + x, y, 1, h))
                        if ii == (length / 2 + length % 2) - 1:
                            y = palette_right_rect.top
                        else:
                            y += color_left_height
                    if self.at_tile_bar:
                        self.label_edit_view.update()
                elif self.at_tile_bar:
                    self.label_new.update()
                if self.at_tile_edit or self.at_palette:
                    if self.at_tile_edit:
                        paint_flashing = True
                        palette_flashing = False
                        mixer_flashing = False
                    elif not self.at_mixer:
                        paint_flashing = False
                        palette_flashing = True
                        mixer_flashing = False
                    else:
                        paint_flashing = False
                        palette_flashing = False
                        mixer_flashing = True
                    self.paint_cursor.set_flashing(paint_flashing)
                    self.paint_cursor.set_size(*([self.CANVAS_SCALE] * 2))
                    self.palette_cursor.set_flashing(palette_flashing)
                    self.palette_cursor.set_size(palette_rect.w, color_left_height)
                    if self.at_mixer or self.is_mixer_selected():
                        self.label_red.location.centery = rect.top + (rect.centery - rect.top) / 2
                        self.label_green.location.centery = rect.centery
                        self.label_blue.location.centery = rect.bottom - (rect.bottom - rect.centery) / 2
                        self.mixer_cursor.set_flashing(mixer_flashing)
                        labels = self.label_red, self.label_green, self.label_blue
                        active_label = labels[self.mixer_index]
                        for label in labels:
                            label.update()
                        self.mixer_cursor.set_size(*active_label.location.size)
                        self.mixer_cursor.set_frame()
                        self.mixer_cursor.location.center = active_label.location.center
                        self.mixer_cursor.update()
                    self.paint_cursor.location.topleft = \
                        rect.left + self.brush_position.x * self.CANVAS_SCALE, \
                        rect.top + self.brush_position.y * self.CANVAS_SCALE
                    if not self.at_mixer and not self.is_mixer_selected():
                        self.paint_cursor.update()
                    if self.at_tile_edit:
                        self.label_paint.update()
                    elif self.at_palette:
                        if self.at_mixer:
                            self.label_add_color.update()
                        elif self.is_mixer_selected():
                            self.label_mix_color.update()
                        else:
                            self.label_select.update()
                    if self.palette_index < length / 2 + length % 2:
                        left = palette_left_rect.left
                        top = color_left_height * self.palette_index
                    else:
                        left = palette_right_rect.left
                        top = color_left_height * (self.palette_index - (length / 2 + length % 2))
                    self.palette_cursor.location.topleft = left - 1, palette_rect.top + top - 1
                    self.palette_cursor.update()
                if self.at_tile_load:
                    self.tile_load_background.update()
                    # self.tile_load_preview_background.update()
                    self.draw_tile_database()
                    self.label_select.update()
                    self.tile_load_cursor.move(-1, -1)
                    self.tile_load_cursor.update()
                    self.tile_load_cursor.move(1, 1)
            if self.at_box:
                if self.box.location.centery > ds.get_rect().centery:
                    self.label_drag.location.top = self.VIEW_MARGIN
                    self.label_back_box.location.top = self.VIEW_MARGIN
                    self.label_switch_tile.location.bottom = self.label_back_box.location.top - 10
                else:
                    self.label_drag.location.bottom = ds.get_height() - self.VIEW_MARGIN
                    self.label_back_box.location.bottom = ds.get_height() - self.VIEW_MARGIN
                    self.label_switch_tile.location.top = self.label_back_box.location.bottom + 10
                surface = Surface(self.box.location.size)
                tile = self.get_current_level().tiles[self.tile_index]
                for x in range(0, surface.get_width(), self.get_level_tile_size()):
                    for y in range(0, surface.get_height(), self.get_level_tile_size()):
                        surface.blit(tile, (x, y))
                surface.set_alpha(self.BOX_PREVIEW_ALPHA)
                ds.blit(surface, self.box.location.topleft)
                self.box.move(-1, -1)
                self.box.update()
                self.box.move(1, 1)
                self.label_drag.update()
                self.label_back_box.update()
                self.label_switch_tile.update()
            if self.at_test:
                interface = self.get_game().interface
                if interface.closed:
                    self.label_back.update()
                    self.label_swap.update()
                else:
                    if self.get_current_level().is_solved():
                        self.label_solved.update()
                interface.update()

    def is_mixer_selected(self):
        return self.palette_index == len(self.get_full_palette())

    def get_full_palette(self):
        return self.default_palette + self.get_current_colors() + self.added_colors

    def get_current_tile(self):
        return self.get_current_level().original_tiles[self.tile_index]

    def get_current_colors(self):
        colors = []
        if not self.add_tile_selected():
            tile = self.get_current_tile()
            for x in range(0, tile.get_width()):
                for y in range(0, tile.get_height()):
                    color = tile.get_at((x, y))
                    if color not in colors and color not in self.default_palette and color not in self.added_colors:
                        colors.append(color)
        return colors


class MixableColor(Color):

    def __init__(self, r=None, g=None, b=None):
        if r is None:
            r = 255
        if g is None:
            g = 0
        if b is None:
            b = 0
        Color.__init__(self, r, g, b)

    def shift_hsla(self, dh=0, ds=0, dl=0, da=0, wrap=False):
        h, s, l, a = self.hsla
        if wrap:
            self.hsla = int(h + dh) % 360, int(s + ds) % 101, int(l + dl) % 101, int(a + da) % 101
        else:
            h += dh
            s += ds
            l += dl
            a += da
            if h < 0:
                h = 0
            elif h > 360:
                h = 360
            if s < 0:
                s = 0
            elif s > 100:
                s = 100
            if l < 0:
                l = 0
            elif l > 100:
                l = 100
            if a < 0:
                a = 0
            elif a > 100:
                a = 100
            self.hsla = h, s, l, a

    def shift_rgb(self, r=0, g=0, b=0, wrap=False):
        if wrap:
            self.r = (self.r + r) % 255
            self.g = (self.g + g) % 255
            self.b = (self.b + b) % 255
        else:
            red = self.r + r
            green = self.g + g
            blue = self.b + b
            if red < 0:
                red = 0
            elif red > 255:
                red = 255
            if green < 0:
                green = 0
            elif green > 255:
                green = 255
            if blue < 0:
                blue = 0
            elif blue > 255:
                blue = 255
            self.r = red
            self.g = green
            self.b = blue


class FlashingCursor(Sprite):

    THICKNESS = 3

    def __init__(self, parent, width=0, height=0, flashing=True):
        Sprite.__init__(self, parent)
        self.set_flashing(flashing)
        self.set_size(width, height)
        self.hue = 0
        self.thickness = self.THICKNESS
        self.set_frame()

    def set_flashing(self, state):
        self.flashing = state

    def set_size(self, width, height):
        self.size = Vector(width, height)

    def set_frame(self):
        offset = self.thickness / 2
        surface = Surface((self.size.x + offset * 2, self.size.y + offset * 2))
        rect = offset, offset, self.size.x, self.size.y
        if self.flashing:
            self.hue = (self.hue + 30) % 360
            surface.set_colorkey((0, 0, 0))
            color = Color(0, 0, 0)
            color.hsla = self.hue, 100, 50, 100
        else:
            surface.fill((255, 0, 0))
            surface.set_colorkey((255, 0, 0))
            color = 0, 0, 0
        draw.rect(surface, color, rect, self.thickness)
        draw.rect(surface, (255, 255, 255), rect, 1)
        self.clear_frames()
        self.add_frame(surface)

    def grow(self, dx=0, dy=0):
        self.size += dx, dy

    def update(self):
        self.set_frame()
        Sprite.update(self)


class SoundEffects(GameChild):

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        effects = self.effects = []
        for path in glob(join(self.get_resource("aud/effect"), "*.wav")):
            effects.append(SoundEffect(self, path))

    def play(self, name, volume=None):
        for effect in self.effects:
            if effect.name == name:
                effect.play(volume=volume)


class SoundEffect(GameChild, Sound):

    DEFAULT_VOLUME = .5

    def __init__(self, parent, path, volume=DEFAULT_VOLUME):
        GameChild.__init__(self, parent)
        Sound.__init__(self, path)
        self.name = basename(path).split(".")[0]
        if self.name == "swap":
            volume = .4
        elif self.name == "menu-move":
            volume = .22
        elif self.name == "cancel":
            volume = .35
        elif self.name == "start":
            volume = .35
        elif self.name == "clear":
            volume = .35
        self.display_surface = self.get_display_surface()
        self.set_volume(volume)

    def play(self, loops=0, maxtime=0, fade_ms=0, position=None, x=None, volume=None):
        channel = Sound.play(self, loops, maxtime, fade_ms)
        if x is not None:
            position = float(x) / self.display_surface.get_width()
        if position is not None and channel is not None:
            channel.set_volume(*self.get_panning(position))
        return channel

    def get_panning(self, position):
        return 1 - max(0, ((position - .5) * 2)), \
               1 + min(0, ((position - .5) * 2))


class Glyphs(GameChild):

    TILE_SIZE = 8
    TRANSPARENT_COLOR = (255, 0, 255)
    LETTER_COLOR = (255, 255, 255)

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        sheet = load(self.get_resource("font.png")).convert()
        sheet.set_colorkey(sheet.get_at((0, 0)))
        original = self.original_tiles = []
        tiles = self.tiles = []
        for y in range(0, sheet.get_height(), self.TILE_SIZE):
            surface = Surface([self.TILE_SIZE] * 2)
            surface.set_colorkey(self.TRANSPARENT_COLOR)
            surface.fill(self.TRANSPARENT_COLOR)
            surface.blit(sheet, (0, -y))
            original.append(surface)
            surface = scale(surface, [surface.get_width() * \
                                      PictureProcessing.SCALE] * 2)
            tiles.append(surface)

    def get_surface_from_text(self, text, color=None, spacing=1,
                              key=TRANSPARENT_COLOR, background=None):
        tw, th = self.tiles[0].get_size()
        tiles = self.get_tiles(text, color)
        surface = Surface((len(tiles) * tw + spacing * len(tiles), th))
        if background is None:
            surface.set_colorkey(key)
            surface.fill(key)
        else:
            surface.fill(background)
        for ii, x in enumerate(range(0, surface.get_width(), tw + spacing)):
            surface.blit(tiles[ii], (x, 0))
        return surface

    def get_tiles(self, text, color):
        tiles = []
        index = 0
        while index < len(text):
            if text[index] == "\\":
                index += 1
                if text[index] == "\\":
                    tile_index = ord(text[index])
                    index += 1
                else:
                    code = ""
                    while index < len(text) and text[index].isdigit():
                        code += text[index]
                        index += 1
                    tile_index = int(code)
            else:
                tile_index = ord(text[index])
                index += 1
            tiles.append(self.get_tile(tile_index, color))
        return tiles

    def get_tile(self, index, color=None):
        tile = self.tiles[index]
        if color is not None:
            tile = tile.copy()
            pixels = PixelArray(tile)
            pixels.replace(self.LETTER_COLOR, color)
            del pixels
        return tile


class Levels(GameChild):

    CLIP_SETTINGS = ((.325, 5512), (.25, 11025), (.5, 11025), (.325, 5512), (.5, 11025))
    TITLES = "POWER UP", "SPIN OUT", "SELECTION", "SPOOKY DIG", "DON'T FALL MOUSE"

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        levels = self.levels = []
        for ii, directory in enumerate(sorted(glob(join(self.get_resource("tiles"),
                                                        "[0-9]*")))):
            level = Level(self)
            level.load(Level.LOAD_MAP_FILE, directory, self.TITLES[ii])
            levels.append(level)

    def reset(self):
        self.deactivate()
        self.current_level = self.levels[0]

    def deactivate(self):
        self.active = False
        self.stop_audio()
        self.get_game().interface.deactivate()

    def activate(self):
        self.active = True

    def stop_audio(self):
        for level in self.levels:
            level.stop_all_audio()

    def get_current_level_index(self):
        return self.levels.index(self.current_level)

    def is_final_level(self):
        return self.get_current_level_index() == len(self.levels) - 1

    def begin_next_level(self):
        self.get_game().interface.deactivate()
        self.levels[self.get_current_level_index() + 1].begin()

    def get_current_title(self):
        return self.TITLES[self.get_current_level_index()]

    def update(self):
        if self.active:
            self.current_level.update()


class Level(Animation):

    VOLUME = .5
    TITLE_OFFSET = 32
    TITLE_BORDER_SIZE = 20, 8
    TITLE_BORDER_SPEED = 400
    TITLE_LENGTH = 5000
    DEFAULT_LEVEL_TITLE = "DEFAULT"
    LOAD_DEFAULT, LOAD_GRID_FILE, LOAD_MAP_FILE = range(3)
    PREVIEW_SCALE = 2

    def __init__(self, parent):
        Animation.__init__(self, parent)
        self.at_title = False
        self.clip_audio = None
        self.full_audio = None
        self.title_plate = None
        self.path = None

    def load(self, method=LOAD_DEFAULT, directory=None, title=None):
        self.grid = []
        self.tiles = []
        self.original_tiles = []
        self.colored_tiles = {}
        if method == self.LOAD_GRID_FILE:
            loaded = self.load_grid_file(directory)
        elif method == self.LOAD_MAP_FILE:
            loaded = self.load_map_file(directory, title)
        else:
            loaded = self.load_default()
        if loaded:
            self.set_preview()
        return loaded

    def load_map_file(self, directory, title):
        tiles = self.tiles
        original_tiles = self.original_tiles
        ds = self.get_display_surface()
        for path in sorted(glob(join(directory, "*.png"))):
            self.add_tile(load(path).convert_alpha())
        colored = self.colored_tiles
        self.set_entire_grid()
        block = 0
        for line in open(join(directory, "map")):
            if line.strip() == "":
                block += 1
            elif block == 0:
                fields = line.split("-")
                if len(fields) == 1:
                    self.default_tile_index = int(fields[0])
                else:
                    for field in fields:
                        digits = [int(member) for member in field.split()]
                        # digits = map(int, field.split())
                        if len(digits) == 1:
                            index = digits[0]
                        elif len(digits) == 3:
                            surface = Surface(tiles[index].get_size())
                            surface.fill(digits)
                            surface.blit(tiles[index], (0, 0))
                            colored[index] = surface
                        else:
                            if len(digits) == 2:
                                digits *= 2
                            for x in range(digits[0], digits[2] + 1):
                                for y in range(digits[1], digits[3] + 1):
                                    self.grid[x][y] = index
        for x, row in enumerate(self.grid):
            for y, value in enumerate(row):
                if value is None:
                    self.grid[x][y] = self.default_tile_index
        title_plate = self.title_plate = Sprite(self)
        title_plate.add_frame(self.get_game().glyphs.get_surface_from_text(title, background=(0, 0, 0)))
        title_background = self.title_background = Surface(ds.get_size())
        title_background.fill((0, 0, 0))
        scene_plate = self.scene_plate = Sprite(self)
        level_number = (Levels.TITLES.index(title) + 1)
        if level_number == len(Levels.TITLES):
            scene_text = "- FINAL SCENE -"
        else:
            scene_text = "- SCENE %d -" % level_number
        scene_plate.add_frame(self.get_game().glyphs.get_surface_from_text(scene_text, background=(0, 0, 0)))
        dsr = ds.get_rect()
        scene_plate.location.center = dsr.centerx, dsr.centery - self.TITLE_OFFSET
        border_tiles = self.border_tiles = []
        for ii in range(self.TITLE_BORDER_SIZE[0] * 2 + self.TITLE_BORDER_SIZE[1] * 2 - 4):
            border_tiles.append(self.tiles[ii % len(self.tiles)])
        path = glob(join(self.get_resource("aud/level"), basename(directory) + "*"))[0]
        self.full_audio = Sound(path)
        self.full_audio.set_volume(self.VOLUME)
        self.title_border_offset = 0
        self.register(self.rotate_title_border, interval=self.TITLE_BORDER_SPEED)
        self.play(self.rotate_title_border)
        self.subscribe(self.respond)
        self.set_swap_status()
        self.path = directory
        return True

    def add_tile(self, surface):
        self.original_tiles.append(surface)
        scaled = scale(surface, [surface.get_width() * PictureProcessing.SCALE] * 2)
        self.tiles.append(scaled)
        self.set_swap_status()

    def get_visible_tile_indicies(self):
        if not self.grid:
            indicies = set((0,))
        else:
            indicies = set()
            for row in self.grid:
                for index in row:
                    indicies.add(index)
        return indicies

    def set_entire_grid(self, index=None):
        self.grid = []
        ds = self.get_display_surface()
        for x in range(int(ds.get_width() / self.tiles[0].get_width())):
            self.grid.append([index] *
                             (int(ds.get_height() / self.tiles[0].get_height())))

    def load_grid_file(self, directory):
        for path in sorted(glob(join(directory, "*.png"))):
            self.add_tile(load(path).convert())
        if not self.tiles:
            return False
        level_path = join(
            directory, self.get_configuration("editor", "level-file-name"))
        if not exists(level_path):
            return False
        level_file = open(level_path, "r")
        self.set_entire_grid()
        for jj, line in enumerate(level_file):
            for ii, character in enumerate(line.strip()):
                self.grid[ii][jj] = int(character, 16)
                if self.grid[ii][jj] >= len(self.tiles):
                    return False
        self.set_swap_status()
        self.path = directory
        return True

    def load_default(self):
        path = self.get_game().editor.get_default_tile_path()
        if exists(path):
            default = load(path).convert()
        else:
            default = Surface([Editor.TILE_SIZE] * 2)
            default.fill((255, 255, 255))
            for y in range(default.get_height()):
                for x in range([1, 2, 2, 2, 2, 2, 7, 8][y]):
                    default.set_at((x, y), (200, 200, 200))
        self.add_tile(default)
        self.set_entire_grid(0)
        self.set_swap_status()
        return True

    def set_preview(self):
        tw, th = self.original_tiles[0].get_size()
        preview = self.preview = Surface(
            (tw * len(self.grid) * self.PREVIEW_SCALE,
             th * len(self.grid[0]) * self.PREVIEW_SCALE))
        for grid_x, surface_x in enumerate(range(0, preview.get_width(),
                                                  tw * self.PREVIEW_SCALE)):
            for grid_y, surface_y in enumerate(range(0, preview.get_height(),
                                                      th * self.PREVIEW_SCALE)):
                scaled = scale(self.original_tiles[self.grid[grid_x][grid_y]],
                               [tw * self.PREVIEW_SCALE] * 2)
                preview.blit(scaled, (surface_x, surface_y))

    def rotate_title_border(self):
        self.title_border_offset -= 1

    def respond(self, event):
        if self.parent.active and self.parent.current_level == self and self.at_title and \
           self.get_game().delegate.compare(event, "advance") and not self.get_game().slides.active:
            self.close_title()

    def close_title(self):
        self.at_title = False
        self.get_game().interface.setup()
        self.get_game().sound_effects.play("granted")

    def begin(self):
        self.parent.activate()
        self.parent.current_level = self
        if self.full_audio is not None:
            self.set_audio()
        self.set_swap_status(True)
        self.parent.stop_audio()
        if self.clip_audio is not None:
            self.clip_audio.play(-1)
        if self.title_plate is not None:
            self.at_title = True
            self.title_plate.set_alpha(255)
            dsr = self.get_display_surface().get_rect()
            self.title_plate.location.center = dsr.centerx, dsr.centery + self.TITLE_OFFSET
            self.title_plate.unhide()
            self.title_elapsed = 0
        else:
            self.close_title()

    def set_audio(self):
        if version.vernum < (1, 9, 2):
            raw = self.full_audio.get_buffer().raw
        else:
            raw = self.full_audio.get_raw()
        samples = array("h", raw)
        volume, count = Levels.CLIP_SETTINGS[self.parent.get_current_level_index()]
        index = randrange(0, len(samples) - count)
        buf = array("h", [0] * 50000) + \
              samples[index:index + count] + \
              array("h", [0] * 150000)
        if version.vernum < (1, 9, 2):
            clip = self.clip_audio = Sound(buf)
        else:
            clip = self.clip_audio = Sound(buffer=buf)
        clip.set_volume(volume)

    def set_swap_status(self, randomize=False):
        swap_status = self.swap_status = {}
        if randomize and len(self.get_visible_tile_indicies()) > 1:
            swapped = False
            indicies = self.get_visible_tile_indicies()
            while not swapped:
                available_swap_positions = indicies.copy()
                for index in indicies:
                    if len(available_swap_positions) == 1 and \
                       list(available_swap_positions)[0] == index:
                        break
                    while True:
                        swap_status[index] = choice(list(available_swap_positions))
                        if swap_status[index] != index:
                            break
                    available_swap_positions.remove(swap_status[index])
                    if not available_swap_positions:
                        swapped = True
                        break
        else:
            for index in self.get_visible_tile_indicies():
                swap_status[index] = index

    def stop_all_audio(self):
        if self.clip_audio:
            self.clip_audio.stop()
        if self.full_audio:
            self.full_audio.stop()

    def is_solved(self):
        return all(index == position for index, position in self.swap_status.items())

    def get_remaining_swap_count(self):
        off = 0
        for index, position in self.swap_status.items():
            off += index != position
        return off / 2 + (off % 2)

    def update(self):
        Animation.update(self)
        ds = self.get_display_surface()
        if self.at_title:
            ds.blit(self.title_background, (0, 0))
            bw, bh = self.TITLE_BORDER_SIZE
            tr = self.tiles[0].get_rect()
            br = Rect(0, 0, bw * tr.w, bh * tr.h)
            br.center = ds.get_rect().center
            index = self.title_border_offset % len(self.border_tiles)
            y = br.top
            for x in range(br.left, br.right, tr.w):
                ds.blit(self.border_tiles[index % len(self.border_tiles)], (x, y))
                index += 1
            for y in range(br.top + tr.h, br.bottom, tr.h):
                ds.blit(self.border_tiles[index % len(self.border_tiles)], (x, y))
                index += 1
            for x in range(br.right - tr.w * 2, br.left - tr.w, -tr.w):
                ds.blit(self.border_tiles[index % len(self.border_tiles)], (x, y))
                index += 1
            for y in range(br.bottom - tr.h * 2, br.top, -tr.h):
                ds.blit(self.border_tiles[index % len(self.border_tiles)], (x, y))
                index += 1
            self.scene_plate.update()
            self.title_plate.update()
            self.title_elapsed += self.get_game().time_filter.get_last_frame_duration()
            if self.title_elapsed >= self.TITLE_LENGTH:
                self.close_title()
        else:
            for grid_x, screen_x in enumerate(range(0, ds.get_width(),
                                          self.tiles[0].get_width())):
                for grid_y, screen_y in enumerate(range(0, ds.get_height(),
                                              self.tiles[0].get_height())):
                    tile_index = self.grid[grid_x][grid_y]
                    swap_index = self.swap_status[tile_index]
                    if swap_index in self.colored_tiles.keys():
                        tile = self.colored_tiles[swap_index]
                    else:
                        tile = self.tiles[swap_index]
                    ds.blit(tile, (screen_x, screen_y))


class Interface(Animation):

    BLINK_INTERVAL = 400
    BLINK_COUNTDOWN_INTERVAL = 2400
    OPEN_TEXT = "ENTER: EDIT MEMORY", "A: EDIT MEMORY"
    CELL_ALPHA = 240
    OPEN_TEXT_OFFSET = 36
    MAX_TILE_COUNT = 16
    CELL_INDENT = 32
    CELL_MARGIN = 56
    CELL_VERTICAL_PADDING = 24
    BACKGROUND = 0, 0, 0
    TEXT_COLOR = 255, 255, 255
    NAME_CHARACTER = "H"
    CELL_INDICATOR = 16
    CELL_INDICATOR_OFFSET = 2
    SHADOW_COLOR = 255, 255, 255
    SHADOW_OFFSET = -1, 1
    SWAP_INDICATOR = 16
    ADVANCE_TEXT = "PRESS ENTER", "PRESS START"
    SWAP_TEXT = "ENT: SWAP", "A: SWAP"
    HIDE_TEXT = "ESC: VIEW", "B: VIEW"
    MOVE_TEXT = ": MOVE"
    PLAY_SONG_DELAY = 1000
    UNSUPPRESS_ADVANCE_DELAY = 2000
    MOVE_ICONS = 24, 25, 26, 27
    COUNTDOWN_THRESHOLD = 2
    TILE_ZOOM_SCALE = 2

    def __init__(self, parent):
        Animation.__init__(self, parent)
        ds = self.display_surface = self.get_display_surface()
        glyphs = self.get_game().glyphs
        open_plate = self.open_plate = Sprite(self)
        open_plate.add_frame(glyphs.get_surface_from_text(self.get_game().select_text(self.OPEN_TEXT),
                                                          background=(0, 0, 0)))
        open_plate.location.midbottom = ds.get_rect().centerx, ds.get_height() - \
                                        self.OPEN_TEXT_OFFSET
        self.cell_frames = cell_frames = []
        tile_size = glyphs.get_tile(0).get_width()
        for ii in range(self.MAX_TILE_COUNT):
            title = ("%02x%s" % (ii, self.NAME_CHARACTER)).upper()
            title_surface = glyphs.get_surface_from_text(title, self.TEXT_COLOR)
            frame = Surface((self.CELL_INDENT * 2 + title_surface.get_width() + \
                             self.CELL_MARGIN + tile_size,
                             self.CELL_VERTICAL_PADDING * 2 + tile_size))
            frame.fill(self.BACKGROUND)
            frame.blit(title_surface, (self.CELL_INDENT, self.CELL_VERTICAL_PADDING))
            cell_frames.append(Sprite(self))
            cell_frames[-1].add_frame(frame)
            cell_frames[-1].set_alpha(self.CELL_ALPHA)
        indicator = self.cell_indicator = Sprite(self)
        indicator.add_frame(glyphs.get_tile(self.CELL_INDICATOR, self.TEXT_COLOR))
        indicator = self.swap_indicator = Sprite(self)
        indicator.add_frame(glyphs.get_tile(self.SWAP_INDICATOR, self.TEXT_COLOR))
        shadow = self.tile_shadow = Surface([tile_size * self.TILE_ZOOM_SCALE] * 2)
        shadow.fill(self.SHADOW_COLOR)
        advance_plate = self.advance_plate = Sprite(self)
        advance_plate.add_frame(glyphs.get_surface_from_text(self.get_game().select_text(self.ADVANCE_TEXT),
                                                             background=(0, 0, 0)))
        advance_plate.location.midbottom = open_plate.location.midbottom
        swap_plate = self.swap_plate = Sprite(self)
        text = self.SWAP_TEXT[1] 
        swap_plate.add_frame(glyphs.get_surface_from_text(self.get_game().select_text(self.SWAP_TEXT),
                                                          background=(0, 0, 0)))
        swap_plate.set_alpha(self.CELL_ALPHA)
        move_plate = self.move_plate = Sprite(self)
        text_surface = glyphs.get_surface_from_text(self.MOVE_TEXT,
                                                    background=(0, 0, 0))
        frame = Surface((tile_size * len(self.MOVE_ICONS) + text_surface.get_width(),
                         tile_size))
        x = 0
        for index in self.MOVE_ICONS:
            frame.blit(glyphs.get_tile(index), (x, 0))
            x += tile_size
        frame.blit(text_surface, (x, 0))
        move_plate.add_frame(frame)
        move_plate.set_alpha(self.CELL_ALPHA)
        move_plate.location.midtop = ds.get_width() / 2, 0
        hide_plate = self.hide_plate = Sprite(self)
        hide_plate.add_frame(glyphs.get_surface_from_text(self.get_game().select_text(self.HIDE_TEXT),
                                                          background=(0, 0, 0)))
        hide_plate.set_alpha(self.CELL_ALPHA)
        hide_plate.location.topright = ds.get_width(), 0
        self.register(self.blink, interval=self.BLINK_INTERVAL)
        self.register(self.blink_countdown, interval=self.BLINK_COUNTDOWN_INTERVAL)
        self.register(self.play_song)
        self.register(self.unsuppress_advance)
        self.subscribe(self.respond)

    def blink(self):
        if self.closed and not self.solved:
            self.open_plate.toggle_hidden()
        if not self.closed and self.swapping_index is not None:
            self.swap_indicator.toggle_hidden()
        if self.solved and not self.suppressing_commands and not self.get_level().at_title:
            self.advance_plate.toggle_hidden()

    def get_level(self):
        if self.get_game().editor.active:
            level = self.get_game().editor.get_current_level()
        else:
            level = self.get_game().levels.current_level
        return level

    def blink_countdown(self):
        if self.is_countdown_active():
            self.countdown_plate.toggle_hidden()
            if self.countdown_plate.is_hidden():
                self.set_title_plate_visibility(True)
            else:
                self.set_title_plate_visibility(False)

    def set_title_plate_visibility(self, visible=True):
        title_plate = self.get_title_plate()
        if title_plate is not None:
            if visible:
                title_plate.unhide()
            else:
                title_plate.hide()

    def is_countdown_active(self):
        return 0 < self.get_level().get_remaining_swap_count() <= self.COUNTDOWN_THRESHOLD

    def respond(self, event, suppress_sound=False):
        if self.active and not self.suppressing_commands and not self.get_game().slides.active:
            delegate = self.get_game().delegate
            effects = self.get_game().sound_effects
            if self.closed and not self.solved and delegate.compare(event, "action"):
                effects.play("memory")
                self.unclose()
            elif self.closed and self.solved and (delegate.compare(event, "advance") or \
               delegate.compare(event, "action")):
                if not self.get_game().editor.active:
                    effects.play("go")
                    self.advance_plate.hide()
                    if self.get_game().levels.is_final_level():
                        self.get_game().ending.activate()
                        self.get_game().levels.deactivate()
                    else:
                        self.get_game().levels.begin_next_level()
            elif not self.closed:
                if delegate.compare(event, "cancel"):
                    effects.play("cancel")
                    if self.swapping_index is not None:
                        self.swapping_index = None
                    else:
                        self.close()
                elif delegate.compare(event, "action"):
                    if self.swapping_index is None:
                        effects.play("start")
                        self.swapping_index = self.get_selected_tile_index()
                        self.set_indicator_position(False)
                        self.respond(Event(delegate.command_event_id,
                                           {"command": "down", "cancel": False}), True)
                    else:
                        if self.get_selected_tile_index() == self.swapping_index:
                            effects.play("cancel")
                        else:
                            level = self.get_level()
                            status = level.swap_status
                            temp = status[self.swapping_index]
                            selected_index = self.get_selected_tile_index()
                            status[self.swapping_index] = status[selected_index]
                            status[selected_index] = temp
                            self.swap_count += 1
                            if not level.is_solved():
                                effects.play("swap")
                                self.set_countdown_plate()
                                if self.is_countdown_active():
                                    self.countdown_plate.unhide()
                                    self.set_title_plate_visibility(False)
                                    self.reset_timer(self.blink_countdown)
                                else:
                                    self.set_title_plate_visibility(True)
                                    self.countdown_plate.hide()
                            elif not self.get_game().editor.active:
                                effects.play("clear")
                                if not self.get_game().levels.is_final_level():
                                    self.get_game().write_progress(self.get_game().levels.\
                                                                   get_current_level_index() + 1,
                                                                   self.swap_count, True)
                                else:
                                    self.get_game().ending.\
                                        set_swap_count(self.get_game().get_progress()[1] + \
                                                       self.swap_count)
                                    self.get_game().write_progress(0, 0)
                                self.suppressing_commands = True
                                self.solved = True
                                self.play(self.play_song, delay=self.PLAY_SONG_DELAY, play_once=True)
                        self.swapping_index = None
                elif delegate.compare(event, ["up", "down", "left", "right"]):
                    if not suppress_sound:
                        effects.play("menu-move")
                    cell_count = self.get_cell_count()
                    bp = int((cell_count - 1) / 2)
                    if delegate.compare(event, "up"):
                        self.cell_index -= 1
                        if self.cell_index == -1:
                            self.cell_index = bp
                        elif self.cell_index == bp:
                            self.cell_index = cell_count - 1
                    elif delegate.compare(event, "down"):
                        self.cell_index += 1
                        if self.cell_index == bp + 1:
                            self.cell_index = 0
                        elif self.cell_index == cell_count:
                            self.cell_index = bp + 1
                    elif delegate.compare(event, "left"):
                        if self.cell_index == 0:
                            self.cell_index = bp + 1
                        elif self.cell_index > bp:
                            self.cell_index -= bp + 1
                        else:
                            self.cell_index += bp
                            if self.cell_index == cell_count:
                                self.cell_index -= 1
                    elif delegate.compare(event, "right"):
                        if self.cell_index == cell_count - 1:
                            self.cell_index = bp
                        elif self.cell_index <= bp:
                            self.cell_index += bp + 1
                            if self.cell_index == cell_count:
                                self.cell_index -= 1
                        else:
                            self.cell_index -= bp
                    self.set_indicator_position()

    def get_selected_tile_index(self):
        for ii, index in enumerate(sorted(self.get_level().swap_status.keys())):
            if ii == self.cell_index:
                return index
                    
    def play_song(self):
        self.close()
        level = self.get_level()
        level.clip_audio.stop()
        level.full_audio.play(-1)
        self.play(self.unsuppress_advance, delay=self.UNSUPPRESS_ADVANCE_DELAY,
                  play_once=True)

    def unsuppress_advance(self):
        self.suppressing_commands = False
        self.advance_plate.unhide()

    def set_indicator_position(self, cell=True):
        if cell:
            indicator = self.cell_indicator
        else:
            indicator = self.swap_indicator
        indicator.location.midleft = self.cell_frames[self.cell_index].location.\
                                     midleft
        indicator.location.left += self.CELL_INDICATOR_OFFSET

    def get_cell_count(self):
        return len(self.get_level().swap_status)

    def reset(self):
        self.deactivate()
        self.solved = False
        self.close()
        self.halt()
        self.suppressing_commands = False

    def deactivate(self):
        self.active = False

    def setup(self):
        self.activate()
        self.close()
        cell_count = self.get_cell_count()
        left_count = int(cell_count / 2 + cell_count % 2)
        right_count = int(cell_count / 2)
        sw, sh = self.display_surface.get_size()
        step = int(sh / (left_count + 1))
        frames = self.cell_frames
        index = 0
        for y in range(step, step * (left_count + 1), step):
            frames[index].location.midleft = 0, y
            index += 1
        step = int(sh / (right_count + 1))
        for y in range(step, step * (right_count + 1), step):
            frames[index].location.midright = sw, y
            index += 1
        self.cell_index = 0
        self.set_indicator_position()
        self.swapping_index = None
        self.swap_count = 0
        self.suppressing_commands = False
        self.solved = False
        self.advance_plate.hide()
        self.play(self.blink)
        self.play(self.blink_countdown)
        dsr = self.get_display_surface().get_rect()
        title_plate = self.get_title_plate()
        if title_plate is not None:
            title_plate.location.midbottom = dsr.centerx, dsr.h
        self.set_countdown_plate()
        self.countdown_plate.hide()

    def get_title_plate(self):
        return self.get_level().title_plate

    def set_countdown_plate(self):
        countdown_plate = self.countdown_plate = Sprite(self)
        count = self.get_level().get_remaining_swap_count()
        swap_text = "SWAPS" if count > 1 else "SWAP"
        text = "%i %s UNTIL FIXED!" % (count, swap_text)
        countdown_plate.add_frame(self.get_game().glyphs.
                                  get_surface_from_text(text, background=(0, 0, 0)))
        countdown_plate.location.midbottom = self.get_display_surface().get_rect().midbottom

    def activate(self):
        self.active = True

    def close(self):
        self.closed = True
        if not self.solved:
            self.open_plate.unhide()

    def unclose(self):
        self.closed = False
        self.open_plate.hide()

    def update(self):
        Animation.update(self)
        if self.active:
            if not self.get_game().editor.active:
                self.open_plate.update()
            self.advance_plate.update()
            if not self.closed:
                for frame in self.cell_frames[:self.get_cell_count()]:
                    frame.update()
                self.cell_indicator.update()
                if self.swapping_index is not None:
                    self.swap_indicator.update()
                level = self.get_level()
                for ii, index in enumerate(sorted(level.swap_status.keys())):
                    position = level.swap_status[index]
                    if position in level.colored_tiles.keys():
                        tile = level.colored_tiles[position]
                    else:
                        tile = level.tiles[position]
                    zoomed = scale(tile, [tile.get_width() * self.TILE_ZOOM_SCALE] * 2)
                    rect = zoomed.get_rect()
                    frame = self.cell_frames[ii]
                    rect.midright = frame.location.right - self.CELL_INDENT, \
                                    frame.location.centery
                    self.display_surface.blit(self.tile_shadow, rect.move(self.SHADOW_OFFSET))
                    self.display_surface.blit(zoomed, rect)
                self.swap_plate.update()
                self.hide_plate.update()
                self.move_plate.update()
                title_plate = self.get_title_plate()
                if title_plate is not None:
                    title_plate.update()
                if not self.get_game().editor.active:
                    self.countdown_plate.update()


class Title(Animation):

    BLINK_INTERVAL = 400
    MENU_OFFSET = 100
    MENU_WIDTH = 440
    ADVANCE_TEXT = "PRESS ENTER", "PRESS START"
    MENU_OPTIONS = "NEW GAME", "CONTINUE", "EDITOR"
    MENU_MARGIN = 12
    MENU_PADDING = 16
    TRANSPARENT_COLOR = (128, 128, 128)
    BACKGROUND_FRAME_COUNT = 1
    BACKGROUND_FRAMERATE = 200
    MENU_INDICATOR_LOCATION = 224
    CHUNK_PROBABILITY = .5
    CHUNK_VOLUME = .1
    NAME = "PICTURE PROCESSING"
    NAME_OFFSET = 300

    def __init__(self, parent):
        Animation.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        self.set_audio()
        self.set_name_plate()
        self.set_advance_plate()
        self.register(self.blink, interval=self.BLINK_INTERVAL)
        self.subscribe(self.respond)
        self.play(self.blink)

    def set_audio(self):
        self.audio_loop = Sound(self.get_resource("aud/title.wav"))
        chunks = self.audio_chunks = []
        for path in glob(join(self.get_resource("aud/chunk"), "*.ogg")):
            chunks.append(Sound(path))
            chunks[-1].set_volume(self.CHUNK_VOLUME)

    def set_name_plate(self):
        plate = self.name_plate = RainbowSprite(
            self, self.get_game().glyphs.get_surface_from_text(self.NAME, background=(80, 0, 0)))
        plate.location.center = self.get_display_surface().get_rect().centerx, self.NAME_OFFSET

    def set_advance_plate(self):
        offset = 0
        plate = self.advance_plate = Sprite(self, self.BLINK_INTERVAL)
        plate.add_frame(self.get_game().glyphs.\
                        get_surface_from_text(self.get_game().select_text(self.ADVANCE_TEXT),
                                              (255, 255, 255), background=(0, 0, 0)))
        rect = self.display_surface.get_rect()
        plate.location.center = rect.centerx, rect.bottom - self.MENU_OFFSET

    def blink(self):
        self.advance_plate.toggle_hidden()
        
    def reset(self):
        self.set_background()
        self.activate()
        self.advance_pressed = False
        self.advance_plate.unhide()
        self.background.set_scroll_step()
        self.menu_index = 0

    def set_background(self):
        self.background = ScrollingBackground(self,
            count=self.BACKGROUND_FRAME_COUNT,
            framerate=self.BACKGROUND_FRAMERATE)

    def activate(self):
        self.active = True
        # self.audio_loop.play(-1)
        self.set_menu()

    def set_menu(self):
        glyphs = self.get_game().glyphs
        if self.get_game().get_progress()[0] > 0:
            option_texts = self.MENU_OPTIONS
        else:
            option_texts = [self.MENU_OPTIONS[0], self.MENU_OPTIONS[2]]
        if not self.get_configuration("editor", "enabled"):
            option_texts = option_texts[:-1]
        height = glyphs.get_tile(0).get_height() * len(option_texts) + \
                 self.MENU_MARGIN * (len(option_texts) - 1) + \
                 self.MENU_PADDING * 2 + 3
        background = self.menu_background = Surface((int(self.MENU_WIDTH),
                                                     height))
        background.fill((0, 0, 0))
        rect = self.menu_rect = background.get_rect()
        rect.center = self.advance_plate.location.center
        options = self.options = []
        for option in option_texts:
            options.append(Sprite(self))
            options[-1].add_frame(glyphs.get_surface_from_text(option))
            options[-1].location.centerx = rect.centerx
            options[-1].location.top = rect.top + self.MENU_PADDING + \
                                       (len(options) - 1) * options[0].location.h + \
                                       (len(options) - 1) * self.MENU_MARGIN
        indicator = self.indicator = Sprite(self)
        indicator.add_frame(self.get_game().glyphs.get_tile(16))
        self.option_texts = option_texts

    def respond(self, event):
        if self.active and not self.get_game().slides.active:
            delegate = self.get_game().delegate
            if not self.advance_pressed:
                if delegate.compare(event, "advance") or delegate.compare(event, "action"):
                    self.advance_pressed = True
                    self.get_game().sound_effects.play("start")
                elif delegate.compare(event, "cancel"):
                    self.get_game().sound_effects.play("cancel")
            else:
                if delegate.compare(event, "cancel"):
                    self.get_game().sound_effects.play("cancel")
                    self.advance_pressed = False
                elif delegate.compare(event, "up"):
                    self.get_game().sound_effects.play("menu-move")
                    self.menu_index -= 1
                elif delegate.compare(event, "down"):
                    self.get_game().sound_effects.play("menu-move")
                    self.menu_index += 1
                elif delegate.compare(event, "advance") or delegate.compare(event, "action"):
                    self.get_game().sound_effects.play("start")
                    if self.get_selected_option() == "NEW GAME":
                        self.deactivate()
                        self.get_game().introduction.load()
                        self.get_game().write_progress(0, 0)
                    elif self.get_selected_option() == "CONTINUE":
                        self.deactivate()
                        self.get_game().levels.\
                            levels[self.get_game().get_progress()[0]].begin()
                    elif self.get_selected_option() == "EDITOR":
                        self.deactivate()
                        self.get_game().editor.activate()
                    else:
                        pass
                    event.dict["command"] = ""
                if self.menu_index < 0:
                    self.menu_index = len(self.options) - 1
                elif self.menu_index >= len(self.options):
                    self.menu_index = 0

    def get_selected_option(self):
        return self.option_texts[self.menu_index]

    def deactivate(self):
        self.active = False
        self.audio_loop.stop()

    def update(self):
        Animation.update(self)
        if self.active:
            self.background.update()
            if not self.advance_pressed:
                self.advance_plate.update()
            else:
                self.display_surface.blit(self.menu_background, self.menu_rect)
                for option in self.options:
                    option.update()
                self.indicator.location.centery = self.options[self.menu_index]\
                                                      .location.centery
                self.indicator.location.right = self.MENU_INDICATOR_LOCATION
                self.indicator.update()
            if random() < self.CHUNK_PROBABILITY and \
               get_busy_channel_count() < get_num_channels() - 5:
                for _ in range(randint(1, 4)):
                    chunk = choice(self.audio_chunks)
                    channel = chunk.play()
                    if channel is not None and not isinstance(channel, tuple):
                        channel.set_volume(random(), random())
            self.name_plate.update()


class ScrollingBackground(Sprite):

    DEFAULT_SIZE = 40 * 24, 40 * 24
    DEFAULT_SCROLL_SPEED = 2
    TURN_PROBABILITY = .0001

    def __init__(self, parent, tiles=None, size=DEFAULT_SIZE, fill=(0, 0, 0),
                 count=1, framerate=None):
        Sprite.__init__(self, parent, framerate)
        frames = []
        for _ in range(count):
            frames.append(Surface((size[0], size[1])))
            frames[-1].fill(fill)
        if tiles is None:
            tiles = self.get_all_tiles()
        tw = tiles[0].get_width()
        fw, fh = frames[0].get_size()
        ox = -(tw - (fw % tw)) / 2 if fw % tw else 0
        oy = -(tw - (fh % tw)) / 2 if fh % tw else 0
        for x in range(ox, fw - ox, tw):
            for y in range(oy, fh - oy, tw):
                for frame in frames:
                    tile = choice(tiles)
                    if random() >= .5:
                        inverted = Surface((tw, tw))
                        inverted.fill((255, 255, 255))
                        inverted.blit(tile, (0, 0), None, BLEND_SUB)
                        tile = inverted
                    fx, fy = random() >= .5, random() >= .5
                    if fx or fy:
                        tile = flip(tile, fx, fy)
                    frame.blit(tile, (x, y))
        for frame in frames:
            self.add_frame(frame)
        self.add_location(offset=(self.location.w, 0))
        self.add_location(offset=(0, self.location.h))
        self.add_location(offset=(self.location.w, self.location.h))
        self.set_scroll_step()

    def get_all_tiles(self):
        tiles = []
        for level in self.get_game().levels.levels:
            tiles.extend(level.tiles)
        return tiles

    def set_scroll_step(self, speed=DEFAULT_SCROLL_SPEED):
        self.scroll_step = get_delta(randrange(0, 360), speed)

    def update(self):
        if random() < self.TURN_PROBABILITY:
            self.set_scroll_step()
        self.move(*self.scroll_step)
        if self.location.top > 0:
            self.move(dy=-self.location.h)
        if self.location.right < 0:
            self.move(self.location.w)
        if self.location.bottom < 0:
            self.move(dy=self.location.h)
        if self.location.left > 0:
            self.move(-self.location.w)
        Sprite.update(self)


class Introduction(Animation):

    BACKGROUND = (0, 0, 0)
    TEXT_COLOR = (255, 255, 255)
    INDICATOR_INDEX = 31
    TEXT = "ONE AFTERNOON IN YOUR/LIVING ROOM", "HOOKING UP YOUR TV GAME/SYSTEM", \
           "PLAYING CARTRIDGES FROM/YOUR COLLECTION", "SUDDENLY THE GRAPHICS/GLITCH UNCONTROLLABLY", \
           "MEDITATING YOU ACCESS/THE CPU TELEPATHICALLY", "FIXING GAMES BY SWAPPING/TILES IN MEMORY"
    INDENT = 48
    OFFSET = 128
    MARGIN = 12
    BLINK_INTERVAL = 400
    PANEL_MARGIN = 100

    def __init__(self, parent):
        Animation.__init__(self, parent)
        ds = self.display_surface = self.get_display_surface()
        background = self.background = Surface(ds.get_size())
        background.fill(self.BACKGROUND)
        blocks = self.blocks = []
        glyphs = self.get_game().glyphs
        tile_height = glyphs.get_tile(0).get_height()
        for block in self.TEXT:
            blocks.append([])
            for ii, line in enumerate(block.split("/")):
                plate = Sprite(self)
                plate.add_frame(glyphs.\
                                get_surface_from_text(line, self.TEXT_COLOR,
                                                      background=self.BACKGROUND))
                plate.location.left = self.INDENT
                plate.location.top = ds.get_height() - self.OFFSET + ii * \
                                     (tile_height + self.MARGIN)
                blocks[-1].append(plate)
        indicator = self.indicator = Sprite(self)
        indicator.add_frame(glyphs.get_tile(self.INDICATOR_INDEX, self.TEXT_COLOR))
        indicator.location.top = ds.get_height() - self.OFFSET + 2 * \
                                 (tile_height + self.MARGIN)
        indicator.location.right = ds.get_width() - self.INDENT
        panels = self.panels = []
        for path in sorted(glob(join(self.get_resource("intro"), "*.png"))):
            panels.append(Sprite(self))
            panels[-1].load_from_path(path)
            panels[-1].location.midtop = ds.get_rect().centerx, self.PANEL_MARGIN
        self.register(self.blink, interval=self.BLINK_INTERVAL)
        self.subscribe(self.respond)
        self.play(self.blink)

    def respond(self, event):
        if self.active and not self.get_game().slides.active:
            if self.get_game().delegate.compare(event, "action"):
                self.block_index += 1
                self.current_panel_index += 1
                if self.block_index == len(self.blocks):
                    self.get_game().sound_effects.play("go")
                    # self.get_game().sound_effects.play("granted")
                    self.deactivate()
                    self.get_game().levels.levels[0].begin()
                # elif self.block_index == len(self.blocks) - 1:
                    # self.get_game().sound_effects.play("granted")
                else:
                    self.get_game().sound_effects.play("next")

    def deactivate(self):
        self.active = False

    def load(self):
        self.activate()
        self.block_index = 0
        self.current_panel_index = 0
        self.indicator.unhide()

    def activate(self):
        self.active = True

    def blink(self):
        self.indicator.toggle_hidden()

    def reset(self):
        self.deactivate()

    def update(self):
        Animation.update(self)
        if self.active:
            ds = self.display_surface
            ds.blit(self.background, (0, 0))
            for block in self.blocks[self.block_index]:
                block.update()
            self.indicator.update()
            self.panels[self.current_panel_index].update()


class Ending(Animation):

    GLYPH_RANGE = 1, 31
    SCROLL_SPEED = 2
    TURN_PROBABILITY = .0001
    PLATE_SIZE = 312, 72
    PLATE_BACKGROUND = 0, 0, 0
    TEXT_COLOR = 255, 255, 255
    END_TEXT = "THE END!"
    SWAP_COUNT_TEXT = "SWAPS:"
    UNSUPPRESS_DELAY = 1000

    def __init__(self, parent):
        Animation.__init__(self, parent)
        ds = self.display_surface = self.get_display_surface()
        end_plate = self.end_plate = Sprite(self)
        frame = Surface(self.PLATE_SIZE)
        frame.fill(self.PLATE_BACKGROUND)
        text_surface = self.get_game().glyphs.get_surface_from_text(self.END_TEXT,
                                                                    self.TEXT_COLOR)
        rect = text_surface.get_rect()
        rect.center = frame.get_rect().center
        frame.blit(text_surface, rect)
        end_plate.add_frame(frame)
        # end_plate.location.center = ds.get_rect().centerx, 3 * ds.get_height() / 8
        end_plate.location.center = ds.get_rect().center
        self.audio = Sound(self.get_resource("aud/end.ogg"))
        self.audio.set_volume(Level.VOLUME)
        self.subscribe(self.respond)
        self.register(self.unsuppress_advance)

    def respond(self, event):
        if self.active and not self.suppress_commands and not self.get_game().slides.active and \
           (self.get_game().delegate.compare(event, "advance") or \
            self.get_game().delegate.compare(event, "action")):
            self.deactivate()
            self.get_game().reset()

    def unsuppress_advance(self):
        self.suppress_commands = False

    def set_swap_count(self, count):
        self.swap_count = count

    def reset(self):
        self.deactivate()
        self.halt()

    def deactivate(self):
        self.active = False
        self.audio.stop()

    def activate(self):
        self.active = True
        self.set_swap_plate()
        self.set_background()
        self.set_scroll_step()
        self.suppress_commands = True
        self.play(self.unsuppress_advance, delay=self.UNSUPPRESS_DELAY,
                  play_once=True)
        self.audio.play(-1)

    def set_swap_plate(self):
        swap_count_plate = self.swap_count_plate = Sprite(self)
        frame = Surface(self.PLATE_SIZE)
        frame.fill(self.PLATE_BACKGROUND)
        text_surface = self.get_game().glyphs.\
                       get_surface_from_text("%s %i" % (self.SWAP_COUNT_TEXT,
                                                        self.swap_count), self.TEXT_COLOR)
        rect = text_surface.get_rect()
        rect.center = frame.get_rect().center
        frame.blit(text_surface, rect)
        swap_count_plate.add_frame(frame)
        swap_count_plate.location.center = self.display_surface.get_rect().centerx, \
                                           5 * self.display_surface.get_height() / 8

    def set_background(self):
        background = self.background = Sprite(self)
        glyphs = self.get_game().glyphs
        tw = glyphs.get_tile(0).get_width()
        ds = self.get_display_surface()
        frame = Surface(ds.get_size())
        for xi, x in enumerate(range(0, ds.get_width(), tw)):
            for yi, y in enumerate(range(0, ds.get_height(), tw)):
                background_color = Color(0, 0, 0)
                background_color.hsla = randrange(0, 360), 100, 50, 100
                foreground_color = Color(0, 0, 0)
                foreground_color.hsla = (background_color.hsla[0] + 180) % 360, \
                                        100, 50, 100
                surface = Surface([tw] * 2)
                surface.fill(background_color)
                surface.blit(glyphs.get_tile(randint(*self.GLYPH_RANGE)), (0, 0))
                frame.blit(surface, (x, y))
        background.add_frame(frame)
        background.add_location(offset=(background.location.w, 0))
        background.add_location(offset=(0, background.location.h))
        background.add_location(offset=(background.location.w,
                                        background.location.h))
                
    def set_scroll_step(self):
        self.scroll_step = get_delta(randrange(0, 360), self.SCROLL_SPEED)

    def update(self):
        Animation.update(self)
        if self.active:
            if random() < self.TURN_PROBABILITY:
                self.set_scroll_step()
            self.background.move(*self.scroll_step)
            if self.background.location.top > 0:
                self.background.move(dy=-self.background.location.h)
            if self.background.location.right < 0:
                self.background.move(self.background.location.w)
            if self.background.location.bottom < 0:
                self.background.move(dy=self.background.location.h)
            if self.background.location.left > 0:
                self.background.move(-self.background.location.w)
            self.background.update()
            self.end_plate.update()
            # self.swap_count_plate.update()
